#include "ofApp.h"
#include "Layer.h"
#include "LayerComposition.h"
#include "MsdfLayer.h"
#include "conversion.h"
#include "import-font.h"
#include "of3dUtils.h"
#include "ofAppRunner.h"
#include "ofColor.h"
#include "ofEvents.h"
#include "ofFileUtils.h"
#include "ofGraphics.h"
#include "ofGraphicsBaseTypes.h"
#include "ofGraphicsConstants.h"
#include "ofTexture.h"
#include "ofUtils.h"
#include "ofWindowSettings.h"
#include "ofxProfiler.h"
#include <GL/glext.h>
#include <filesystem>
#ifdef TARGET_EMSCRIPTEN
 #include <emscripten/em_asm.h>
 #include <GLES/gl.h>
 #include <GLES3/gl3.h>
#endif
#include <memory>

static int AA = 1;

namespace VariableEditor {
//--------------------------------------------------------------
void ofApp::setup(){
    OFX_PROFILER_FUNCTION();
    EM_ASM({window.setLoadingTask('setting up rendering', 0)});

    {
        ofFile sf("appSettings.json");
        if(sf.exists()){
            ofJson json = ofLoadJson("appSettings.json");
            settings.from_json(json, settings);
        }
    }

    ofSetFrameRate(30);
    ofSetVerticalSync(false);

    cam.setVFlip(true); // otherwise everything is weird
    cam.setPosition(ofGetWidth() / 2, ofGetHeight() / 2, 935.335);
    //cam.setPosition(0, 0, -1000);
    cam.lookAt(glm::vec3(cam.getPosition().x, cam.getPosition().y, 0),
               glm::vec3(0, 1, 0));
    //cam.disableMouseInput();
    cam.enableOrtho();

    observer.setupPerspective(true, // vflip
                              60, // fov
                              0, // nearDist
                              10000000, // farDist
                              glm::vec2(0, 0) // lensOffset
                              );
    EM_ASM({window.setLoadingTask('setting up rendering', 10)});

    ofDisableArbTex();
    fboSettings.width = ofGetWidth() * AA;
    fboSettings.height = ofGetHeight() * AA;
    fboSettings.numSamples = 0;
    fboSettings.internalformat = GL_RGBA;
    fboSettings.minFilter = GL_LINEAR_MIPMAP_LINEAR;
    fboSettings.maxFilter = GL_LINEAR;
    //fboSettings.textureTarget = GL_TEXTURE_2D_MULTISAMPLE;
    fbo.allocate(fboSettings);

    fbo.begin();
    ofClear(ofColor(settings.backgroundColor[0],
                    settings.backgroundColor[1],
                    settings.backgroundColor[2],
                    settings.backgroundColor[3]));
    fbo.end();

    //fbo.allocate(ofGetWidth() * AA, ofGetHeight() * AA, GL_RGB);

    EM_ASM({window.setLoadingTask('setting up rendering', 30)});
    layerComposition.setup();
    EM_ASM({window.setLoadingTask('setting up rendering', 90)});
    layerComposition.setVFlip(true);

    #ifndef TARGET_OPENGLES
        {
            std::string fontPath = "data/fonts/Version-2-var.ttf";
            ofxVariableLab::LayerType type = ofxVariableLab::LayerType::GPUFONT;
            ofxVariableLab::Layer::Props props;
            props.fontPath = fontPath;
            props.text = "yo, whatever you want, and especially pancakes";
            props.y = 120;
            props.x = 95;
            layerComposition.addLayer(
                {fontPath, type},
                props,
                {{"Weight", 100.0}, {"Weight", 700.0}}
                );
        }
    #endif

    image42.load("42px-01.png");
    image420.load("420px-01.png");

    int maxSamples;
    glGetIntegerv(GL_MAX_SAMPLES, &maxSamples);
    cout << "MAX_SAMPLES: " << ofToString(maxSamples) << endl;

    int maxTextureSize;
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxTextureSize);
    cout << "MAX_TEXTURE_SIZE: " << ofToString(maxTextureSize) << endl;
    if(maxTextureSize < 16384){
        #ifdef TARGET_EMSCRIPTEN
            // yeaaaaahh, well this is mostly for firefox on a mac
            EM_ASM({
                alert('Your browser / operating system only allows quite small textures, which can lead to glitches.\n\nWe recommend:\nMac OS + Chrome\nWindows + Chrome\nLinux + Chrome / Firefox\n\nPS: If you do not want to use Chrome, Brave is a good Chrome-based browser.');
            });
        #endif
    }

    const unsigned char * glVersion = glGetString(GL_VERSION);
    int glMajor, glMinor;
    glGetIntegerv(GL_MAJOR_VERSION, &glMajor);
    glGetIntegerv(GL_MINOR_VERSION, &glMinor);

    cout << "GL_VERSION: " << glVersion
         << " | Major(" << ofToString(glMajor) << ")"
         << " | Minor(" << ofToString(glMajor) << ")"
         << endl;
    artboard.setup();
    EM_ASM({window.setLoadingTask('setting up rendering', 100)});
}

//--------------------------------------------------------------
void ofApp::update(){
    OFX_PROFILER_FUNCTION();
    if(!rendering || renderNextFrame){
        layerComposition.update();
    }
    zipSaver.update();
    projectZipSaver.update();
}

//--------------------------------------------------------------
void ofApp::draw(){
    OFX_PROFILER_FUNCTION();
    if(!rendering || renderNextFrame){
        ofCamera & camera = observing ? observer : cam;

        artboard.begin();
        camera.begin();
        ofPushMatrix();
        layerComposition.draw(artboard.getShape().x,
                              artboard.getShape().y);
        ofPopMatrix();
        camera.end();
        artboard.end();

        fbo.begin();
        ofFloatColor bg(settings.backgroundColor[0],
                        settings.backgroundColor[1],
                        settings.backgroundColor[2],
                        settings.backgroundColor[3]);
        ofClear(bg);
        artboard.draw();
        fbo.end();
        ofDisableAlphaBlending();
        ofDisableDepthTest();

        fbo.getTexture().generateMipmap();
        fbo.draw(0, 0, ofGetWidth(), ofGetHeight());

        ofPushStyle();
        ofDrawBitmapStringHighlight("fps: " + ofToString(ofGetFrameRate()), 20, ofGetHeight() - 120);
        //ofDrawBitmapStringHighlight("ortho: " + ofToString(cam.getOrtho()), 20, ofGetHeight() - 200);
        ofPopStyle();

        ofEnableDepthTest();
    }
    if(rendering){
        //if(currentFrame < guessedFrameCount){
        int n_zero = log10(guessedFrameCount) + 1;
        string frame_number_padded = std::string(n_zero, 'x');
        if(currentFrame > 0){
            string frame_number = ofToString(currentFrame - 1);
            frame_number_padded = std::string(n_zero - std::min(n_zero, (int)frame_number.length()), '0') + frame_number;
            saveFrame(frame_number_padded, artboard.getFbo());
            recordedFrameNames.push_back(frame_number_padded);
        }
        currentFrame++;
        #ifdef TARGET_EMSCRIPTEN
            int percent = (theatrePosition / timelineLength_seconds) * 100;
            EM_ASM_INT({
                let percent = $0;
                document.getElementById('export_progress_task').innerHTML = 'rendering';
                let innerHTML = "|";
                const niceText = "rendering is the process of generating an image by means of a computer program. ";
                for(let i = 0; i < 100; i++){
                    if(i < percent){
                        innerHTML += niceText[i % niceText.length];
                    }else{
                        innerHTML += "-";
                    }
                }
                innerHTML += "|";
                let progress = document.getElementById("export_progress");
                progress.innerHTML = innerHTML;
            }, percent);
            double theatrePostPosition = EM_ASM_DOUBLE({
                window.tp.sheet.sequence.position = $0;
                return window.tp.sheet.sequence.position;
            }, theatrePosition);
            if(theatrePostPosition >= timelineLength_seconds
               || theatrePosition > theatrePostPosition){
                cout << std::fixed
                     << "------------------------- "
                     << "frame: " << frame_number_padded << endl
                     << "theatrePosition: " << ofToString(theatrePosition) << std::endl
                     << "timelineLEngth|_second: " << ofToString(timelineLength_seconds) << std::endl
                     << "theatrePostPosition: " << ofToString(theatrePostPosition) << std::endl
                ;
                EM_ASM({
                    window.renderDone();
                });
                theatrePosition = 0;
                renderNextFrame = 0;
                rendering = false;
            }else{
                theatrePosition += timeScale / 30.0f;
                renderNextFrame = true;
            }
        #endif
    }
    //}
}

void ofApp::drawGrid(){

}

void ofApp::setPlaying(bool playing){
    this->playing = playing;
}


//--------------------------------------------------------------
void ofApp::keyPressed(int key){
    //cout << "pressed " << (char)key << "(" << ofToString(key) << ")" << endl;
    if(inputPressed.count(ofKey(key)) == 0){
        inputPressed.insert(ofKey(key));
    }
    return;
    if(key == 's'){
        #ifdef TARGET_EMSCRIPTEN
            //atlasImage->save("web_atlasImage.png");
            for(const auto & a : layerComposition.getAtlasLayerCombos()){
                if(a.second->getIdentifier().type == ofxVariableLab::LayerType::MSDFGEN){
                    auto combo =
                        static_pointer_cast <ofxVariableLab::MsdfAtlasLayerCombo>(a.second);
                    string imageName = combo->getAtlasImagePath();
                    ofStringReplace(imageName, "/", "_");
                    ofStringReplace(imageName, "data_atlascache_", "");
                    downloadImage(imageName, combo->getAtlasImage());
                }else if(a.second->getIdentifier().type == ofxVariableLab::LayerType::GPUFONT){
                    auto combo =
                        static_pointer_cast <ofxVariableLab::GPUFontAtlasLayerCombo>(a.second);
                }
            }
        #else
            for(const auto & a : layerComposition.getAtlasLayerCombos()){
                if(a.second->getIdentifier().type == ofxVariableLab::LayerType::MSDFGEN){
                    auto combo =
                        static_pointer_cast <ofxVariableLab::MsdfAtlasLayerCombo>(a.second);
                    string imageName = combo->getAtlasImagePath();
                    ofStringReplace(imageName, "/", "_");
                    ofStringReplace(imageName, "data_atlascache_", "");
                    combo->getAtlasImage().save(imageName);
                }else if(a.second->getIdentifier().type == ofxVariableLab::LayerType::GPUFONT){
                    auto combo =
                        static_pointer_cast <ofxVariableLab::GPUFontAtlasLayerCombo>(a.second);
                }
            }
        #endif
    }
    if(key == 'a'){
        #ifdef TARGET_EMSCRIPTEN
            downloadFboAsImage("web_fboImage.png", fbo);
            //atlasImage->save("web_atlasImage.png");
        #else
            ofPixels pixels;
            fbo.readToPixels(pixels);
            ofSaveImage(pixels, "linux64_fboImage.png");
        #endif
    }
    if(key == 'o'){
        if(cam.getOrtho()){
            cout << "diable ortho" << endl;
            cam.disableOrtho();
        }else{
            cout << "eable ortho" << endl;
            cam.enableOrtho();
        }
    }
    if(key == 'p'){
        observing = !observing;
    }
    if(key == 'c'){
        cout << "observer to cam" << endl;
        observer.setGlobalOrientation(cam.getGlobalOrientation());
        observer.setGlobalPosition(cam.getGlobalPosition());
        observer.setScale(cam.getGlobalScale());
        observer.setFov(cam.getFov());
    }
}

//--------------------------------------------------------------
void ofApp::keyReleased(int key){
    //cout << "released " << (char)key << "(" << ofToString(key) << ")" << endl;
    if(inputPressed.count(ofKey(key)) > 0){
        inputPressed.erase(ofKey(key));
    }
    #ifdef OFX_PROFILER
        //if(key == 'o'){
        //OFX_PROFILER_BEGIN_SESSION("profiling", "results.json");
        //}
        //if(key == 'p'){
        //OFX_PROFILER_END_SESSION();

        //#ifdef TARGET_EMSCRIPTEN
        //std::ifstream f("results.json");
        //nlohmann::json json = nlohmann::json::parse(f);
        //downloadJson("results.json", json);
        //#endif
        //cout << json.dump() << endl;
        //}
        //if(key == 'r'){
        //rendering = !rendering;
        //}
    #endif
}

//--------------------------------------------------------------
void ofApp::mouseMoved(int x, int y){
    //cout << "mouse moved" << endl;
    //if(inputPressed.count(OF_MOUSE_BUTTON_MIDDLE) > 0
    //|| (inputPressed.count(OF_MOUSE_BUTTON_LEFT) > 0 && inputPressed.count(OF_KEY_CONTROL) > 0)){
    //artboard.setPosition(x, y, 0);
    //}
}

//--------------------------------------------------------------
void ofApp::mouseDragged(int x, int y, int button){
}

//--------------------------------------------------------------
void ofApp::mousePressed(int x, int y, int button){
    //cout << "mouse pressed " << button << endl;
    if(inputPressed.count(ofKey(button)) == 0){
        inputPressed.insert(ofKey(button));
    }

}

//--------------------------------------------------------------
void ofApp::mouseReleased(int x, int y, int button){
    //cout << "mouse released " << button << endl;
    if(inputPressed.count(ofKey(button)) == 0){
        inputPressed.erase(ofKey(button));
    }
}

//--------------------------------------------------------------
void ofApp::mouseEntered(int x, int y){
    //cout << "mouse entered" << endl;

}

//--------------------------------------------------------------
void ofApp::mouseExited(int x, int y){
    //cout << "mouse exited" << endl;

}

void ofApp::mouseScrolled(ofMouseEventArgs & mouse){
    this->mouseScrolled(mouse.x, mouse.y, mouse.scrollX, mouse.scrollY);
}

void ofApp::mouseScrolled(int x, int y, float scrollX, float scrollY){
    //cout << "scroll "
    //<< "x: " << ofToString(x) << " "
    //<< "y: " << ofToString(y) << " "
    //<< "scrollX: " << ofToString(scrollX) << " "
    //<< "scrollY: " << ofToString(scrollY) << " "
    //;
}

//--------------------------------------------------------------
void ofApp::windowResized(int w, int h){
    cout << "window resized " << ofToString(w) << "x" << ofToString(h) << endl;
    ofSetWindowShape(w, h);
    fboSettings.width = w * AA;
    fboSettings.height = h * AA;
    fboSettings.numSamples = 0;
    fboSettings.internalformat = GL_RGBA;
    fboSettings.minFilter = GL_LINEAR_MIPMAP_LINEAR;
    fboSettings.maxFilter = GL_LINEAR;
    //fboSettings.textureTarget = GL_TEXTURE_2D_MULTISAMPLE;
    ofFbo newFbo;
    newFbo.allocate(fboSettings);

    newFbo.begin();
    ofClear(ofColor(settings.backgroundColor[0],
                    settings.backgroundColor[1],
                    settings.backgroundColor[2],
                    settings.backgroundColor[3]));
    newFbo.end();
    fbo = std::move(newFbo);
}

//--------------------------------------------------------------
void ofApp::gotMessage(ofMessage msg){

}

//--------------------------------------------------------------
void ofApp::dragEvent(ofDragInfo dragInfo){

}

void ofApp::saveFrame(const string & filename,
                      const ofFbo & _fbo){
    ofImage image;
    image.setUseTexture(false);
    image.allocate(_fbo.getWidth(),
                   _fbo.getHeight(),
                   OF_IMAGE_COLOR_ALPHA);
    _fbo.readToPixels(image.getPixels());
    image.save(settings.tmpExportDir + "/frames/" + filename + ".png");
}

#ifdef TARGET_EMSCRIPTEN
    void ofApp::downloadImage(const string & filename,
                              const ofImage & image){
        downloadPixelsAsImage(filename, image.getPixels());
    }
    void ofApp::downloadFramesAsZip(const string & projectName){
        if(!zipSaver.isThreadRunning()){
            cout << "ofApp::downloadFramesAsZip" << endl;
            zipSaver.setup(projectName,
                           settings.tmpExportDir + "/frames",
                           recordedFrameNames
                           );
            zipSaver.startThread();
        }
    }

    void ofApp::downloadFboAsImage(const string & filename,
                                   const ofFbo & _fbo){
        ofPixels pixels;
        _fbo.readToPixels(pixels);
        downloadPixelsAsImage(filename, pixels);
    }

    void ofApp::downloadFrame(const string & filename){
        cout << "downloadFrame " << filename << endl;
        ofImage image;
        image.load(settings.tmpExportDir + "/frames/" + filename + ".png");
        downloadImage(filename,
                      image);
    }

    void ofApp::downloadPixelsAsImage(const string & filename,
                                      const ofPixels & pixels){
        ofBuffer buffer;
        ofSaveImage(pixels, buffer, OF_IMAGE_FORMAT_PNG);
        emscripten_browser_file::download(filename.c_str(),
                                          "image/png",
                                          buffer.getData(),
                                          buffer.size());
    }

    void ofApp::downloadJson(const string & filename,
                             const nlohmann::json & json){
        string jsonString = json.dump();
        emscripten_browser_file::download(filename.c_str(),
                                          "application/json",
                                          jsonString.data(),
                                          jsonString.size());
    }

    void ofApp::downloadProject(const string & projectName,
                                const string & projectJsonString){
        if(!projectZipSaver.isThreadRunning()){
            projectZipSaver.setup(projectName,
                                  projectJsonString);
            projectZipSaver.startThread();
        }
    }
#endif
void ofApp::exit(){
    zipSaver.stopThread();
}

}
